/*
 *****************************************************************************
 * Copyright by ams AG                                                       *
 * All rights are reserved.                                                  *
 *                                                                           *
 * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
 * THE SOFTWARE.                                                             *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
 * USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
 * EXCLUDED.                                                                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
 *****************************************************************************
 */


/*! \file
 * \brief Device driver for monitoring ambient light intensity in (lux)
 * proximity detection (prox) functionality within the
 * AMS TMD2755 family of devices.
 */


#include <linux/input.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/types.h>


#include "linux/i2c/ams/ams_tmd2755.h"
#include "ams_tmd2755_prox.h"
#include "ams_i2c.h"

#if defined(INCLUDE_THERMAL_DRIFT_SW_ALGO)
#include "ams_tmd2755_thermal_drift.h"
#endif


/*****************************************************************
 * Local globals for prox module                                 *
 *****************************************************************/

/*****************************************************************
 *  Utility Calls                                                *
 *****************************************************************/
static void tmd2755_read_prox(struct tmd2755_chip *chip)
{
	/* Read low and high bytes */
	ams_i2c_blk_read(chip->client, TMD2755_REG_PDATAL, &chip->shadow[TMD2755_REG_PDATAL], sizeof(chip->prox_info.raw));

	if (chip->params.prox_apc == PROX_APC_ENABLED)
		chip->prox_info.raw = le16_to_cpu(*((const __le16 *)&chip->shadow[TMD2755_REG_PDATAL]));
	else {
		/* APC disabled - PDATAL contains the upper 8 bits and PDATAH 1:0 contains the lower */
		/*   2 bits. */
		chip->prox_info.raw = (u16)((chip->shadow[TMD2755_REG_PDATAL] << 2) | (chip->shadow[TMD2755_REG_PDATAH] & 0x3));
	}

	return;
}

static void tmd2755_get_prox(struct tmd2755_chip *chip)
{
	u16 mask = 0, high_t = 0, low_t = 0;
	u8 l_data[2], h_data[2];

	/* Read thresholds from HW */
	ams_i2c_blk_read(chip->client, TMD2755_REG_PILTL, &l_data[0], 2);
	ams_i2c_blk_read(chip->client, TMD2755_REG_PIHTL, &h_data[0], 2);

	/* Logging thresholds, pdata, state */
	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		low_t  = le16_to_cpu(*((const __le16 *)&l_data[0]));
		high_t = le16_to_cpu(*((const __le16 *)&h_data[0]));
		dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tcurrent state: %s [Raw: %d]  [Min Thresh: %d]  [Max Thresh:%d]  APC: Enabled\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				chip->prox_info.detected ? "<Detect>" : "<Release>", chip->prox_info.raw,
				chip->params.prox_thresh_min, chip->params.prox_thresh_max);

		dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tDevice Register Values: Low Threshold: %d   High Threshold: %d\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				low_t, high_t);
	}
	else {
		low_t  = (l_data[0] & 0xFF); // actual thresholds 10 bits
		high_t = (h_data[0] & 0xFF);
	
		dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tcurrent state: %s [Raw: %d]  [Min Thresh: %d]  [Max Thresh:%d]  APC: Disabled\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				chip->prox_info.detected ? "<Detect>" : "<Release>", chip->prox_info.raw,
				chip->params.prox_thresh_min & TMD2755_MASK_PILTL, chip->params.prox_thresh_max & TMD2755_MASK_PIHTL);
		dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tAdjusted Thresholds(10 bits - lsh2) Data Structures: [Min: %d]   [Max: %d]\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				(chip->params.prox_thresh_min & TMD2755_MASK_PILTL) << 2,
				(chip->params.prox_thresh_max & TMD2755_MASK_PIHTL) << 2);

		dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tDevice Register Values (lsh2): Low Threshold: %d (%d)   High Threshold: %d (%d)\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				low_t, low_t << 2,  high_t, high_t << 2);
	}

	/* Need to use data structure values not register values. Register values   */
	/* are adjusted so that we do not get multiple detect or release interrupts */
	/* We only want one of each.  Therefore, once an interrupt occurs, that     */
	/* threshold is then set to the max in the device register.                 */
	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		low_t  = chip->params.prox_thresh_min;
		high_t = chip->params.prox_thresh_max;
		mask = PROX_THRESH_APC_MAX;
	}
	else {
		low_t  = ((chip->params.prox_thresh_min & 0xFF) << 2);
		high_t = ((chip->params.prox_thresh_max & 0xFF) << 2);
		mask = TMD2755_MASK_PILTL; /* just need any 8 bit mask */
	}

	if (chip->prox_info.detected == false) {
        if (chip->prox_info.raw >= high_t) {
			chip->prox_info.detected = true;
			dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tnew state: <Detect>\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);

			/*If proximity is detected, set the lower limit to find */
			/* proximity release and set the higher limit to avoid  */
			/* repeated interrupts (max val).                       */
			tmd2755_configure_thresholds(chip, chip->params.prox_thresh_min, mask); /* mask is also the max */

			dev_info(&chip->client->dev, "%*.*s():%*d --> New Threshold limits = [%d, %d]\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				chip->params.prox_thresh_min & mask, PROX_THRESH_APC_MAX & mask);
		}
	} else {
		/* Release event - since we are in detected state */
		if (chip->prox_info.raw <= low_t) {
			chip->prox_info.detected = false;
			dev_info(&chip->client->dev, "%*.*s():%*d --> \t\tnew state: <Release>\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);

			/* If proximity is released, set the higher limit to find         */
			/* proximity and set the lower limit to avoid repeated interrupts */
			/* This is same as the init state                                 */
			tmd2755_init_prox_mode(chip);

			dev_info(&chip->client->dev, "%*.*s():%*d --> New Threshold limits = [%d, %d]\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				PROX_THRESH_MIN & mask, chip->params.prox_thresh_max & mask);
		}
	}
	return;
}

/*
 *  Value        Meaning
 *    0.     Prox release event
 *    1.      Prox detect event
 *    2.     Prox saturation event
 */
void tmd2755_prox_report_inp_event(struct tmd2755_chip *chip, int value)
{
	if (chip->prox_idev) {
		input_report_abs(chip->prox_idev, ABS_DISTANCE, value);
		input_sync(chip->prox_idev);
	}
	return;
}

void tmd2755_process_prox_irq(struct tmd2755_chip *chip)
{
	tmd2755_read_prox(chip);
	tmd2755_get_prox(chip);
	tmd2755_prox_report_inp_event(chip, chip->prox_info.detected ? 1 : 0);

	return;
}

void tmd2755_process_saturation_event(struct tmd2755_chip *chip)
{
	tmd2755_prox_report_inp_event(chip, PROX_SAT_EVENT);

	/* Disable PSAT interrupt otherwise the processor will be over-burdened processing */
	/* saturation interrupts                                                           */
	ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_INTENAB,  TMD2755_PSIEN, 0);

	return;
}

static void tmd2755_hw_drift_algo_config(struct tmd2755_chip *chip)
{
	u8 hw_pers = 1;

	/* enable hw drift algorithm */
	ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_DRFT_CNTRL,
					TMD2755_MASK_DRIFT_ENABLE, (1 << TMD2755_SHIFT_DRIFT_ENABLE));

	/* set HW persistance  based on pwtime */
	if (chip->params.prox_wtime == PWTIME_100_MS) /* 35 represents 100 ms */
	{
		hw_pers = HW_DRIFT_ALGO_PERSIST_7;
	}
	else
	{
		hw_pers = HW_DRIFT_ALGO_PERSIST_15;
	}
	
	ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_DRFT_PERSIST,
					TMD2755_MASK_DRIFT_PERSIST, hw_pers << TMD2755_SHIFT_DRIFT_PERSIST);

	/* set the Drift Lower Limit1 - DLL */
	ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_DRFT_DLL, HW_DRIFT_ALGO_DLL);

	/* set the Drift Upper Limit 1 - DUL1 */
	ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_DRFT_DU1L, HW_DRIFT_ALGO_DUL1);

	/* set the Drift Lower Limit - DUL2 */
	ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_DRFT_DU2L, HW_DRIFT_ALGO_DUL2);

	return;
}


void tmd2755_init_prox_mode(struct tmd2755_chip *chip)
{
	/* Init Low threshold to 0 - Only looking for Detect events after init */
	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		dev_info(&chip->client->dev, "%*.*s():%*d --> Setting Initial Thresholds (APC) = [%d, %d]\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				MIN_PROX_THRESHOLD, chip->params.prox_thresh_max);
	}
	else {
		dev_info(&chip->client->dev, "%*.*s():%*d --> Setting Initial Thresholds (No APC) = [%d, %d (%d)]\n",
				MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
				MIN_PROX_THRESHOLD & TMD2755_MASK_PILTL, 
				chip->params.prox_thresh_max & TMD2755_MASK_PIHTL,
				(chip->params.prox_thresh_max & TMD2755_MASK_PIHTL) << 2);
	}
	

	tmd2755_configure_thresholds(chip, MIN_PROX_THRESHOLD, chip->params.prox_thresh_max);

	/* Enable PSAT interrupt during init or after release event */
	ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_INTENAB,  TMD2755_PSIEN, TMD2755_PSIEN);
	return;
}

int tmd2755_offset_calibration(struct tmd2755_chip *chip)
{
	u8 *sh = chip->shadow;
	u8 saveenab, saveint;
	int ret;
	unsigned long wt;

	/* clean up the sync mechanism between here and ISR */
	reinit_completion(&(chip->calibration_done));

	/* save PEN state */
	ams_i2c_read(chip->client, TMD2755_REG_ENABLE, &saveenab);

	/* save prox intr state */
	ams_i2c_read(chip->client, TMD2755_REG_INTENAB, &saveint);

	/* turn on power, disable prox */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_ENABLE, TMD2755_PEN | TMD2755_PON, TMD2755_PON);

	/* enable calib intr, disable prox intr */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_INTENAB, TMD2755_PIEN | TMD2755_CIEN, TMD2755_CIEN);

	/* Clear any interrupts */
	ams_i2c_write_direct(chip->client, TMD2755_REG_STATUS, 0xFF);

	/*
	 ** Prox Offset calibration
	 **   binsrch_target (15 counts)
	 **   prox averaging (2 reading window)
	 **   prox_auto_offset_adjust
	 */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_CALIBCFG,
			TMD2755_MASK_BINSRCH_TARGET | TMD2755_MASK_PROX_AUTO_OFFSET_ADJUST | TMD2755_MASK_PROX_DATA_AVG,
			(chip->params.prox_binsrch_tgt << TMD2755_SHIFT_BINSRCH_TARGET) | (PROX_AUTO_OFFSET_ADJ_VAL << TMD2755_SHIFT_PROX_AUTO_OFFSET_ADJUST) |
			(chip->params.prox_avg << TMD2755_SHIFT_PROX_DATA_AVG));
	/*
	 ** Calibration Offset
	 **   enable offset range extension
	 **   set ORE to PROX_ORE_VAL
	 */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_CALIB_OFF,
			TMD2755_MASK_ENABLE_ORE | TMD2755_MASK_ORE,
			(chip->params.prox_ore << TMD2755_SHIFT_ENABLE_ORE) | (PROX_ORE_VAL << TMD2755_SHIFT_ORE));

	/*
	 * CALAVG: Enable to allow HW averaging
	 * ELEC_CALIB: perform both electrical and optical - default
	 * CALPRATE: Enables PRATE during calibration
	 */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_CALIB, TMD2755_MASK_CALAVG_CALIB | TMD2755_MASK_CALPRATE_CALIB,
			(PROX_HWAVG_CAL << TMD2755_SHIFT_CALAVG_CALIB) |  (PROX_PRATE_CAL << TMD2755_SHIFT_CALPRATE_CALIB));

	/* trigger calibration sequence */
	dev_info(&chip->client->dev, "%*.*s():%*d --> offset calibration started (waiting for completion).\n",
		MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
	chip->amsCalComplete = false;

	chip->in_calib = true;
	ams_i2c_modify(chip->client, sh, TMD2755_REG_CALIB, TMD2755_MASK_START_OFFSET_CALIB, (0x01 << TMD2755_SHIFT_START_OFFSET_CALIB));

	/* wait for ISR to signal calibration int complete */
	AMS_MUTEX_UNLOCK(&chip->lock);
	wt = wait_for_completion_interruptible_timeout(&(chip->calibration_done), msecs_to_jiffies(PROX_CALIB_TIMEOUT_MS));
	AMS_MUTEX_LOCK(&chip->lock);
	if (wt <= 0) {
		dev_info(&chip->client->dev, "%*.*s():%*d --> Proximity Calibration timeout occurred: %lu\n",
			MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,  wt);
		ret = -1;
	} else {
		ret = 0;
		chip->in_calib = false;
		chip->amsCalComplete = true;
		/* get updated prox offset */
		ams_i2c_blk_read(chip->client, TMD2755_REG_POFFSET_L, &chip->shadow[TMD2755_REG_POFFSET_L], 2);
		chip->params.poffset = chip->shadow[TMD2755_REG_POFFSET_L];
		dev_info(&chip->client->dev, "%*.*s():%*d --> Optical Xtalk calib complete(poffsetl=%d, poffseth=%d, ore=%d, ore_en=%d)\n",
			MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, chip->params.poffset,
			chip->shadow[TMD2755_REG_POFFSET_H] & TMD2755_MASK_POFFSET_H, chip->shadow[TMD2755_REG_CALIB_OFF] & TMD2755_MASK_ORE,
			(chip->shadow[TMD2755_REG_CALIB_OFF] & TMD2755_MASK_ENABLE_ORE) > TMD2755_SHIFT_ENABLE_ORE);

		if (chip->shadow[TMD2755_REG_POFFSET_H] & TMD2755_MASK_POFFSET_H)
			chip->params.poffset *= -1;
	}

	/* Restore register enable and interrupt enable */
	ams_i2c_modify(chip->client, sh, TMD2755_REG_ENABLE,  TMD2755_PEN, saveenab);
	ams_i2c_modify(chip->client, sh, TMD2755_REG_INTENAB, TMD2755_PIEN, saveint);

	return ret;
}

int tmd2755_configure_prox_mode(struct tmd2755_chip *chip, u8 state)
{
	struct i2c_client *client = chip->client;
	u8 *sh = chip->shadow;

	dev_info(&chip->client->dev, "%*.*s():%*d --> Configuring Prox Mode %s\n",
		MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, state ? "ON" : "OFF");

	/* Turning on prox */
	if (state) {
		/* Configure default proximity settings - thresholds */
		tmd2755_init_prox_mode(chip);

#if defined(INCLUDE_THERMAL_DRIFT_SW_ALGO)
		ams_tmd2755_sw_drift_setup(chip);
#endif

		/* Enable and configure the HW drift algorithm       */
		tmd2755_hw_drift_algo_config(chip);

		tmd2755_offset_calibration(chip);

		ams_i2c_modify(client, sh, TMD2755_REG_PERS, TMD2755_MASK_PROX_PERS, chip->params.persist.persistance & TMD2755_MASK_PROX_PERS);
		ams_i2c_write(client,  sh, TMD2755_REG_PCFG2, (chip->params.prox_pulse_len  & TMD2755_MASK_PPULSE_LEN_L));
		ams_i2c_modify(client, sh, TMD2755_REG_PCFG1, TMD2755_MASK_REG_PPULSE_LEN_H,
			(chip->params.prox_pulse_len & TMD2755_MASK_DATA_PPULSE_LEN_H) >> TMD2755_SHIFT_PPULSE_LEN_H);

		/* PGAIN2 */
		ams_i2c_modify(client, sh, TMD2755_REG_CFG1,  TMD2755_MASK_PGAIN2, (chip->params.prox_gain2 << TMD2755_SHIFT_PGAIN2));
		/* PGAIN1 */
		ams_i2c_modify(client, sh, TMD2755_REG_PCFG0, TMD2755_MASK_PGAIN1, (chip->params.prox_gain1 << TMD2755_SHIFT_PGAIN1));

		/* PRATE defaults to 0x1F */
		ams_i2c_write(client, sh, TMD2755_REG_PRATE, chip->params.prox_time/* P_TIME_US(INTEGRATION_CYCLE) */);

		/* PWTIME */
		ams_i2c_write(client, sh, TMD2755_REG_PWTIME, chip->params.prox_wtime);

		/* Enable Proximity and Proximity Interrupt */
		/* PWEN is required to be enabled if ALS feature is enabled */
		/* By default, enable PWTIME with Bit 4 in REG_ENABLE */
		ams_i2c_modify(client, sh, TMD2755_REG_ENABLE, TMD2755_PWEN | TMD2755_PEN | TMD2755_PON, TMD2755_PWEN | TMD2755_PEN | TMD2755_PON);
		/* Enable Prox interrupt, saturation interrupt and zero detect interrupt */
		ams_i2c_modify(client, sh, TMD2755_REG_INTENAB, TMD2755_PIEN | TMD2755_PSIEN | TMD2755_ZIEN,
			TMD2755_PIEN | TMD2755_PSIEN | TMD2755_ZIEN);

		chip->prox_enable = true;
		chip->prox_info.detected = 0;
		chip->prox_info.raw = 0;
	} else {
		/* Turning off prox */
		/* Disable Proximity feature, and all interrupts associated with prox */
		ams_i2c_modify(client, sh, TMD2755_REG_ENABLE, TMD2755_PEN, 0);
		ams_i2c_modify(client, sh, TMD2755_REG_INTENAB, TMD2755_PROX_INTS, 0);

#if defined(INCLUDE_THERMAL_DRIFT_SW_ALGO)
		ams_tmd2755_sw_drift_shutdown();
#endif

		/* If nothing else is enabled set PON = 0 */
		if (!(sh[TMD2755_REG_ENABLE] & TMD2755_EN_ALL))
			ams_i2c_modify(client, sh, TMD2755_REG_ENABLE, TMD2755_PON, 0x00);

		chip->prox_enable = false;
	}

	return 0;
}

/*****************************************************************
 *   sysfs utility functions                                     *
 *****************************************************************/
static ssize_t tmd2755_prox_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	u8 idx;
	ssize_t ret = 0;

	for (idx = 0; idx < tmd2755_prox_attrs_size; idx++) {
		if (!strncmp(tmd2755_prox_attrs[idx].attr.name, attr->attr.name, strlen(attr->attr.name))) {
			AMS_MUTEX_LOCK(&chip->lock);
			switch (idx) {
			case PROX_RAW_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n",  chip->prox_info.raw);
				break;
			case PROX_GAIN1_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_gain1);
				break;
			case PROX_GAIN2_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_gain2);
				break;
			case PROX_OFFSET_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.poffset);
				break;
			case PROX_PERSIST_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.persist.pers.ppers);
				break;
			case PROX_PULSE_LEN_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_pulse_len);
				break;
			case PROX_ENABLE_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->prox_enable);
				break;
			case PROX_APC_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_apc);
				break;
			case PROX_TIME_ATTR:
				ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_time);
				break;
			case PROX_LOW_THRESH_ATTR:
			    if (chip->params.prox_apc == PROX_APC_ENABLED)
					ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_thresh_min);
				else {
					/* APC Disabled */
					ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_thresh_min & TMD2755_MASK_PILTL);
				}
				break;
			case PROX_HIGH_THRESH_ATTR:
				if (chip->params.prox_apc == PROX_APC_ENABLED)
					ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_thresh_max);
				else {
					/* APS Disabled */
					ret = scnprintf(buf, PAGE_SIZE, "%d\n", chip->params.prox_thresh_max & TMD2755_MASK_PIHTL);
				}
				break;
			case PROX_SATURATION_ATTR:
				/* Determine if prox is in saturation */
				ret = scnprintf(buf, PAGE_SIZE, "%s\n", 
					(chip->in_psat == PROX_NO_SAT ? "Normal_Operation" : 
					(chip->in_psat == PROX_AMBIENT_SAT ? "Ambient_Saturation" : "Reflective_Saturation")));
				break;
			default:
				ret = scnprintf(buf, PAGE_SIZE, "not found\n");
				break;
			}
			AMS_MUTEX_UNLOCK(&chip->lock);
			break;
		}
	}
	return ret;
}

/*****************************************************************
 *   ABI FUNCTIONS - sysfs callbacks                             *
 *****************************************************************/
static ssize_t tmd2755_prox_raw_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	/* populates the prox_info structure */
	tmd2755_read_prox(chip);

	/* Use scnprintf() */
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_detected_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	u8 val;

	/* if prox INT is enabled, just grab the data flag that is defined in prox_info */
	/* Otherwise read the prox */
	ams_i2c_read(chip->client, TMD2755_REG_INTENAB, &val);
	if (!(val & TMD2755_PIEN)) {
		tmd2755_read_prox(chip);
		tmd2755_get_prox(chip);
	}

	return scnprintf(buf, PAGE_SIZE, "%s\n", chip->prox_info.detected ? "detected" : "released");
}


 /* gains */
 /* stage 1 - */
static ssize_t tmd2755_prox_gain1_show(struct device *dev, struct device_attribute *attr, char *buf)
{

	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_gain1_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long gain;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	int rc;

	rc = kstrtoul(buf, BASE_10, &gain);

	if (rc || (gain > PROX_GAIN_MAX))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();
	rc = ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_PCFG0, TMD2755_MASK_PGAIN1, gain << TMD2755_SHIFT_PGAIN1);
	chip->params.prox_gain1 = gain;
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

/* stage 2 - proximity IR sensor */
static ssize_t tmd2755_prox_gain2_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_gain2_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long gain;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	int rc;

	rc = kstrtoul(buf, BASE_10, &gain);

	if (rc || (gain > PROX_GAIN_MAX) || (gain == 2))   /* 2 is a reserved value for pgain2 */
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();
	rc = ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_CFG1, TMD2755_MASK_PGAIN2, gain << TMD2755_SHIFT_PGAIN2);
	chip->params.prox_gain2 = gain;
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_offset_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_persist_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_persist_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long persist;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	rc = kstrtoul(buf, BASE_10, &persist);
	if (rc || (persist > PROX_PERSIST_MAX))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();
	ams_i2c_modify(chip->client, chip->shadow,
		TMD2755_REG_PERS, TMD2755_MASK_PROX_PERS, persist << TMD2755_SHIFT_PROX_PERS);
	chip->params.persist.persistance = chip->shadow[TMD2755_REG_PERS];
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_pulse_len_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_pulse_len_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long pulse_len;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	rc = kstrtoul(buf, BASE_10, &pulse_len);
	if (rc || (pulse_len < PROX_PULSE_LEN_MIN) || (pulse_len > PROX_PULSE_LEN_MAX))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();

	/* pulse length is 10 bits: 8 bits in PCFG2 and 2 MSBs in PCFG1 */
	ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PCFG2, (pulse_len  & TMD2755_MASK_PPULSE_LEN_L));
	ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_PCFG1, TMD2755_MASK_REG_PPULSE_LEN_H,
		(pulse_len & TMD2755_MASK_DATA_PPULSE_LEN_H) >> TMD2755_SHIFT_PPULSE_LEN_H);
	chip->params.prox_pulse_len = pulse_len;
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_time_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_time_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long prate;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	rc = kstrtoul(buf, BASE_10, &prate);
	if (rc || (prate > PROX_TIME_MAX))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();
	ams_i2c_write(chip->client, chip->shadow,
		TMD2755_REG_PRATE, prate);
	chip->params.prox_time = chip->shadow[TMD2755_REG_PRATE];
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_enable_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_apc_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_enable_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long enable;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	rc = kstrtoul(buf, 10, &enable);
	/* if error on conversion, value greater than 1(true) */
	if (rc || (enable > 1))
		return -EINVAL;
	/* if its the same, no need to do any work. */
	if (enable == chip->prox_enable)
		return size;

	AMS_MUTEX_LOCK(&chip->lock);

	if (enable)
		tmd2755_configure_prox_mode(chip, TMD2755_FEATURE_ON);
	else /* disable */
		tmd2755_configure_prox_mode(chip, TMD2755_FEATURE_OFF);

	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}


/* APC_Disabled is a flag in CFG6: Settings the bit disables APC */
static ssize_t tmd2755_prox_apc_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long apc;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	rc = kstrtoul(buf, 10, &apc);
	/* if error on conversion, value greater than 1(true) */
	if (rc || (apc > 1))
		return -EINVAL;
	/* if its the same, no need to do any work. */
	if (apc == chip->params.prox_apc)
		return size;

	AMS_MUTEX_LOCK(&chip->lock);

	PROX_OFF();
	rc = ams_i2c_modify(chip->client, chip->shadow, TMD2755_REG_CFG6, TMD2755_MASK_APC, apc << TMD2755_SHIFT_APC);
	chip->params.prox_apc = apc;
	PROX_ON();

	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_low_thresh_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_low_thresh_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long low_thresh;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	unsigned long max_low = 0;

	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		max_low = PROX_THRESH_APC_MAX;
	} else {
		max_low = PROX_THRESH_NO_APC_MAX;
	}
	
	rc = kstrtoul(buf, BASE_10, &low_thresh);
	if (rc || (low_thresh > max_low))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();

	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PILTL, (low_thresh & TMD2755_MASK_PILTL));
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PILTH, ((low_thresh >> TMD2755_SHIFT_PIXTH) & TMD2755_MASK_PILTH));
	} else {
		/* apc == 1 means APC is DISABLED, only 8 MSB of PILT are used - threshold is only 14 bit number (i.e. right shift 6) */
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PILTL, (low_thresh & TMD2755_MASK_PILTL));
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PILTH, 0);
	}
	chip->params.prox_thresh_min = low_thresh;

	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_high_thresh_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

static ssize_t tmd2755_prox_high_thresh_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size)
{
	unsigned long high_thresh;
	int rc;
	struct tmd2755_chip *chip = dev_get_drvdata(dev);
	unsigned long max_high;

	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		max_high = PROX_THRESH_APC_MAX;
	} else {
		max_high = PROX_THRESH_NO_APC_MAX;
	}

	rc = kstrtoul(buf, BASE_10, &high_thresh);
	if (rc || (high_thresh > max_high))
		return -EINVAL;

	AMS_MUTEX_LOCK(&chip->lock);
	PROX_OFF();
	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PIHTL, (high_thresh & TMD2755_MASK_PIHTL));
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PIHTH, ((high_thresh >> TMD2755_SHIFT_PIXTH) & TMD2755_MASK_PIHTH));
	} else {
		/* apc == 1 means APC is DISABLED, only 8 MSB of PIHT are used - threshold is only 14 bit number (i.e. right shift 6) */
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PIHTL, (high_thresh & TMD2755_MASK_PIHTL));
		ams_i2c_write(chip->client, chip->shadow, TMD2755_REG_PIHTH, 0);
	}
	chip->params.prox_thresh_max = high_thresh;
	PROX_ON();
	AMS_MUTEX_UNLOCK(&chip->lock);

	return size;
}

static ssize_t tmd2755_prox_regs_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct tmd2755_chip *chip = dev_get_drvdata(dev);

	u8 rEnable  = 0, rPrate  = 0, rPiltl  = 0, rPilth   = 0;
	u8 rPihtl   = 0, rPihth  = 0, rPers   = 0, rCfg0    = 0;
	u8 rPcfg0   = 0, rCfg1   = 0, rStatus = 0, rPcfg1   = 0;
	u8 rCfg8    = 0, rCfg3   = 0, rCfg6   = 0, rIntenab = 0;
	u8 rPwtime  = 0, rCalib  = 0, rCalOff = 0, rCalcfg  = 0;
	u8 rPcfg2   = 0, rCalstat = 0;

	ams_i2c_read(chip->client, TMD2755_REG_ENABLE,  &rEnable);
	ams_i2c_read(chip->client, TMD2755_REG_PRATE,   &rPrate);
	ams_i2c_read(chip->client, TMD2755_REG_PILTL,   &rPiltl);
	ams_i2c_read(chip->client, TMD2755_REG_PILTH,   &rPilth);
	ams_i2c_read(chip->client, TMD2755_REG_PIHTL,   &rPihtl);
	ams_i2c_read(chip->client, TMD2755_REG_PIHTH,   &rPihth);
	ams_i2c_read(chip->client, TMD2755_REG_PERS,    &rPers);

	ams_i2c_read(chip->client, TMD2755_REG_CFG0,    &rCfg0);
	ams_i2c_read(chip->client, TMD2755_REG_CFG1,    &rCfg1);
	ams_i2c_read(chip->client, TMD2755_REG_CFG8,    &rCfg8);
	ams_i2c_read(chip->client, TMD2755_REG_CFG3,    &rCfg3);
	ams_i2c_read(chip->client, TMD2755_REG_CFG6,    &rCfg6);

	ams_i2c_read(chip->client, TMD2755_REG_PCFG0,   &rPcfg0);
	ams_i2c_read(chip->client, TMD2755_REG_PCFG1,   &rPcfg1);
	ams_i2c_read(chip->client, TMD2755_REG_PCFG2,   &rPcfg2);

	ams_i2c_read(chip->client, TMD2755_REG_STATUS,  &rStatus);
	ams_i2c_read(chip->client, TMD2755_REG_PWTIME,  &rPwtime);

	ams_i2c_read(chip->client, TMD2755_REG_INTENAB, &rIntenab);

	/* Calibration Registers */
	ams_i2c_read(chip->client, TMD2755_REG_CALIB,     &rCalib);
	ams_i2c_read(chip->client, TMD2755_REG_CALIB_OFF, &rCalOff);
	ams_i2c_read(chip->client, TMD2755_REG_CALIBCFG,  &rCalcfg);
	ams_i2c_read(chip->client, TMD2755_REG_CALIBSTAT, &rCalstat);


	return scnprintf(buf, PAGE_SIZE,
		"ENABLE   =   0x%02X\nPRATE    =   0x%02X\nPILT     =   0x%04X\nPIHT     =   0x%04X\n"
		"PERS     =   0x%02X\nPCFG0    =   0x%02X\nPCFG1    =   0x%02X\nPCFG2    =   0x%02X\n"
		"CFG0     =   0x%02X\nCFG1     =   0x%02X\nCFG8     =   0x%02X\nCFG3     =   0x%02X\n"
		"CFG6     =   0x%02X\nSTATUS   =   0x%02X\nPWTIME   =   0x%02X\nINTENAB  =   0x%02X\n"
		"CALIB    =   0x%02X\nCAL_OFF  =   0x%02X\nCAL_CFG  =   0x%02X\nCAL_STAT =   0x%02X\n"
		"%s\n",
		rEnable, rPrate, (rPilth << 8) | rPiltl, (rPihth << 8) | rPihtl, rPers,
		rPcfg0, rPcfg1, rPcfg2,
		rCfg0, rCfg1, rCfg8, rCfg3, rCfg6,
		rStatus, rPwtime, rIntenab,
		rCalib, rCalOff, rCalcfg, rCalstat,
		chip->prox_info.detected ? "Prox Detect" : "Prox Release");
}

static ssize_t tmd2755_prox_sat_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return tmd2755_prox_show(dev, attr, buf);
}

/**********************************************************************************************************************
 *  END OF SYSFS FUNCTIONS                                                                                            *
 **********************************************************************************************************************/
/*
 * The prox thresholds are affected by the APC value.  If APC is enabled, the thresholds
 * are 14 bit values.If APC is disabled, the low byte contains the	8 bits that is
 * compared to the upper 8 bits of the 10 bit pdata.  Make sure if APC is disabled,
 * the appropriate thresholdsare used.	The value set in the register is left shifted
 * 2 and compared to pdata.
 */
void tmd2755_configure_thresholds(struct tmd2755_chip *chip, u16 low_t, u16 high_t)
{
	u8 *sh = chip->shadow;

	if (chip->params.prox_apc == PROX_APC_ENABLED) {
		/* Set lower threshold */
		ams_i2c_write(chip->client, sh, TMD2755_REG_PILTL, (u8)(low_t & TMD2755_MASK_PILTL));
		ams_i2c_write(chip->client, sh, TMD2755_REG_PILTH, (u8)((low_t >> TMD2755_SHIFT_PIXTH) & TMD2755_MASK_PILTH));

		/* set the higher threshold */
		ams_i2c_write(chip->client, sh, TMD2755_REG_PIHTL, (u8)(high_t & TMD2755_MASK_PIHTL));
		ams_i2c_write(chip->client, sh, TMD2755_REG_PIHTH, (u8)((high_t >> TMD2755_SHIFT_PIXTH) & TMD2755_MASK_PIHTH));
	} else {
		/* APC is Disabled - These 8 bits are compared to the upper 8 bits of pdata */
		/* basically a lsh of 2 */
		/* Lower */
		ams_i2c_write(chip->client, sh, TMD2755_REG_PILTL, (u8)(low_t & TMD2755_MASK_PILTL));
		ams_i2c_write(chip->client, sh, TMD2755_REG_PILTH, (u8)0);

		/* Upper */
		ams_i2c_write(chip->client, sh, TMD2755_REG_PIHTL, (u8)(high_t & TMD2755_MASK_PIHTL));
		ams_i2c_write(chip->client, sh, TMD2755_REG_PIHTH, (u8)0);
	}

	return;
}

/**************************************************************
 *                   Global Sys Fs definitions                *
 **************************************************************/
struct device_attribute tmd2755_prox_attrs[] = {
	__ATTR(prox_raw,         0444, tmd2755_prox_raw_show,         NULL),
	__ATTR(prox_detect,      0444, tmd2755_prox_detected_show,    NULL),
	__ATTR(prox_gain1,       0644, tmd2755_prox_gain1_show,       tmd2755_prox_gain1_store),
	__ATTR(prox_gain2,       0644, tmd2755_prox_gain2_show,       tmd2755_prox_gain2_store),
	__ATTR(prox_offset,      0444, tmd2755_prox_offset_show,      NULL),
	__ATTR(prox_persist,     0644, tmd2755_prox_persist_show,     tmd2755_prox_persist_store),
	__ATTR(prox_pulse_len,   0644, tmd2755_prox_pulse_len_show,   tmd2755_prox_pulse_len_store),
	__ATTR(prox_enable,      0644, tmd2755_prox_enable_show,      tmd2755_prox_enable_store),
	__ATTR(prox_apc,         0644, tmd2755_prox_apc_show,         tmd2755_prox_apc_store),
	__ATTR(prox_time,        0644, tmd2755_prox_time_show,        tmd2755_prox_time_store),
	__ATTR(prox_low_thresh,  0644, tmd2755_prox_low_thresh_show,  tmd2755_prox_low_thresh_store),
	__ATTR(prox_high_thresh, 0644, tmd2755_prox_high_thresh_show, tmd2755_prox_high_thresh_store),
	__ATTR(prox_regs,        0444, tmd2755_prox_regs_show,        NULL),
	__ATTR(prox_sat,         0444, tmd2755_prox_sat_show,         NULL),
};

int tmd2755_prox_attrs_size = ARRAY_SIZE(tmd2755_prox_attrs);

