/*
 *****************************************************************************
 * Copyright by ams AG                                                       *
 * All rights are reserved.                                                  *
 *                                                                           *
 * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
 * THE SOFTWARE.                                                             *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED FOR USE ONLY IN CONJUNCTION WITH AMS PRODUCTS.  *
 * USE OF THE SOFTWARE IN CONJUNCTION WITH NON-AMS-PRODUCTS IS EXPLICITLY    *
 * EXCLUDED.                                                                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,     *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
 *****************************************************************************
 */

/* THIS SOFTWARE DRIFT ALGORITHM IS BEST USED IN A HIGH CROSSTALK ENVIRONMENT */
/* WHERE APC IS DISABLED                                                      */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/i2c.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/wait.h>

#ifdef CONFIG_OF
#include <linux/of_device.h>
#endif

#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/delay.h>


#include "linux/i2c/ams/ams_tmd2755.h"
#include "ams_i2c.h"
#include "ams_tmd2755_prox.h"

#include "ams_tmd2755_thermal_drift.h"

/* Each array index represents a temperature range bin: */
static s16 thermal_drift_table[DRIFT_TABLE_SIZE] =
{
    DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL,
    DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL,
    DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL,
    DRIFT_TABLE_DEFAULT_VAL, DRIFT_TABLE_DEFAULT_VAL
};

static tmd2755_thermal_drift_info_t drift_info =
{
    .init               = false,
    .chip               = NULL,
    .table_modified     = false,
};

/* Structure specifically used to log drift table data - not essential */
/* to code operation. */
static drift_table_state_t drift_log_info = 
{
    .state         = false,
    .pdata         = 0,
    .poffset       = 0,
    .temp_cnt      = 0,
    .poffset_table = 0,
};

/*
 * Based on the ADC count for the temp sensors, an index for the
 * drift table is returned.  As temperature goes up, the index goes down
 */
static uint16_t get_temp_index(uint16_t count)
{
    uint16_t index = 0;

    if (count <= DRIFT_TEMP_BIN_10_UPPER)
    {
        /* >= 50 C */
        index = 10;
    }
    else if ((count > DRIFT_TEMP_BIN_9_LOWER) && (count <= DRIFT_TEMP_BIN_9_UPPER))
    {
        /* 45 C <= x < 50 C */
        index = 9;
    }
    else if ((count > DRIFT_TEMP_BIN_8_LOWER) && (count <= DRIFT_TEMP_BIN_8_UPPER))
    {
        /* 40 C <= x < 45 C */
        index = 8;
    }
    else if ((count > DRIFT_TEMP_BIN_7_LOWER) && (count <= DRIFT_TEMP_BIN_7_UPPER))
    {
        /* 35 C <= x < 40 C */
        index = 7;
    }
    else if ((count > DRIFT_TEMP_BIN_6_LOWER) && (count <= DRIFT_TEMP_BIN_6_UPPER))
    {
        /* 30 C <= x < 35 C */
        index = 6;
    }
    else if ((count > DRIFT_TEMP_BIN_5_LOWER) && (count <= DRIFT_TEMP_BIN_5_UPPER))
    {
        /* 25 C <= x < 30 C */
        index = 5;
    }
    else if ((count > DRIFT_TEMP_BIN_4_LOWER) && (count <= DRIFT_TEMP_BIN_4_UPPER))
    {
        /* 20 C <= x < 25 C */
        index = 4;
    }
    else if ((count > DRIFT_TEMP_BIN_3_LOWER) && (count <= DRIFT_TEMP_BIN_3_UPPER))
    {
        /* 15 C <= x < 20 C */
        index = 3;
    }
    else if ((count > DRIFT_TEMP_BIN_2_LOWER) && (count <= DRIFT_TEMP_BIN_2_UPPER))
    {
        /* 5 C <= x < 15 C */
        index = 2;
    }
    else if ((count > DRIFT_TEMP_BIN_1_LOWER) && (count <= DRIFT_TEMP_BIN_1_UPPER))
    {
        /* -5 C <= x < 5 C */
        index = 1;
    }
    else if (count >DRIFT_TEMP_BIN_0_LOWER)
    {
        /* < -5 C */
        index = 0;
    }

    return(index);
}

/*
 * Dynamically populate the drift table with the first acquired
 * value.
 */
static void runtime_table_update(uint16_t index)
{
    u8 idx = 0;
    s16 init_val = thermal_drift_table[index];

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Auto-updating Drift Table with first acquired value: %d\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, init_val);
    for (idx = 0; idx < DRIFT_TABLE_SIZE; idx++)
    {
        thermal_drift_table[idx] = init_val;
    }

    return;
}

static void set_drift_table(u16 therm_adc_cnt, s16 poffset)
{
    uint16_t idx = get_temp_index(therm_adc_cnt);

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Replacing index [%d]: From: %d To: %d\n",
                 MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN,
                 __LINE__, idx, thermal_drift_table[idx], poffset);

    thermal_drift_table[idx] = poffset;

    drift_log_info.poffset_table = thermal_drift_table[idx];
    drift_log_info.table_index = idx;

    /* when we receive the first table update, populate the rest of the */
    /* table with the value                                             */
    if (!drift_info.table_modified)
    {
        drift_info.table_modified = true;
        runtime_table_update(idx);
    }

    return;
}

static void init_drift_table(s16 poffset)
{
    u16 idx = 0;

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Initialize Drift Table to First Calibrated poffset: [%d]\n",
            MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, poffset);

    for (idx = 0; idx < DRIFT_TABLE_SIZE; idx++)
    {
        thermal_drift_table[idx] = poffset;
    }

    return;
}

static void custom_prox_config(void)
{
    /* number of pulses */
    drift_info.chip->params.prox_pulse_cnt = CUSTOM_PULSE_CNT;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_PCFG0, TMD2755_MASK_PPULSE,
                   drift_info.chip->params.prox_pulse_cnt);

    /* pulse length */
    drift_info.chip->params.prox_pulse_len = CUSTOM_PULSE_LEN;
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow,
                  TMD2755_REG_PCFG2,
                  drift_info.chip->params.prox_pulse_len & TMD2755_MASK_PPULSE_LEN_L);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_PCFG1, TMD2755_MASK_REG_PPULSE_LEN_H,
                   drift_info.chip->params.prox_pulse_len);

    /* pgain1 */
    drift_info.chip->params.prox_gain1 = CUSTOM_PGAIN1;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_PCFG0, TMD2755_MASK_PGAIN1,
                   (drift_info.chip->params.prox_gain1 << TMD2755_SHIFT_PGAIN1));

    /* gpain2 */
    drift_info.chip->params.prox_gain2 = CUSTOM_PGAIN2;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CFG1, TMD2755_MASK_PGAIN2,
                   (drift_info.chip->params.prox_gain2 << TMD2755_SHIFT_PGAIN2));

    /* pldrive */
    drift_info.chip->params.prox_drive = CUSTOM_PLDRIVE;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_PCFG1, TMD2755_MASK_PLDRIVE,
                   drift_info.chip->params.prox_drive);

    /* prate (ptime) */
    drift_info.chip->params.prox_time = CUSTOM_PRATE;
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow,
                  TMD2755_REG_PRATE, drift_info.chip->params.prox_time);

    /* pwen */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_ENABLE, TMD2755_PWEN, CUSTOM_PWEN);

    /* pwtime */
    drift_info.chip->params.prox_wtime = CUSTOM_PWTIME;
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow,
                  TMD2755_REG_PWTIME, drift_info.chip->params.prox_wtime);

    /* binsrch */
    drift_info.chip->params.prox_binsrch_tgt= CUSTOM_BIN_TGT_SRCH;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CALIBCFG, TMD2755_MASK_BINSRCH_TARGET,
                   (drift_info.chip->params.prox_binsrch_tgt << TMD2755_SHIFT_BINSRCH_TARGET));

    /* ORE */
    drift_info.chip->params.prox_ore = CUSTOM_ORE;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CALIB_OFF, TMD2755_MASK_ENABLE_ORE,
                   (drift_info.chip->params.prox_ore << TMD2755_SHIFT_ENABLE_ORE));

    /* APC_Disable */
    drift_info.chip->params.prox_apc = CUSTOM_APC_DISABLE;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CFG6, TMD2755_MASK_APC,
                   (drift_info.chip->params.prox_apc << TMD2755_SHIFT_APC));

    /* HW Average */
    drift_info.chip->params.prox_avg = CUSTOM_HW_AVG;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CALIBCFG, TMD2755_MASK_HW_AVG,
                   (drift_info.chip->params.prox_avg << TMD2755_SHIFT_HW_AVG));

    /* Detect Threshold */
    /* Release Threshold  - Set to 0, we start in release mode but populate */
    /* data structure with actual value.  This will be populated after a */
    /* detect event. */
    drift_info.chip->params.prox_thresh_max = CUSTOM_DET_THRESH;
    drift_info.chip->params.prox_thresh_min = CUSTOM_REL_THRESH;
    tmd2755_configure_thresholds(drift_info.chip, 0, drift_info.chip->params.prox_thresh_max);

    return;
}

s16 get_drift_table_entry(u16 therm_adc_cnt)
{
    u16 idx = 0;
    s16 poffset = 0;

    idx = get_temp_index(therm_adc_cnt);

    drift_log_info.table_index = idx;

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Requesting poffset for Table index [%d]\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, idx);
    poffset = thermal_drift_table[idx];

    return(poffset);
}

/*
 *  This routine is used to read the pdata registers.  The pdata registers
 *  not onlyu stores the prox count, but is also used to store the Temperature
 *  ADC count.
 */
static u16 acquire_current_pdata(void)
{
    u16 pdata;

    ams_i2c_blk_read(drift_info.chip->client, TMD2755_REG_PDATAL,
                     &drift_info.chip->shadow[TMD2755_REG_PDATAL], 2);

    /* Currently, custom_prox_config() diabled APC - so this could be unnecessary */
    /* but the custom config can change from custiomer to customer */
    if (drift_info.chip->params.prox_apc == PROX_APC_ENABLED)
    {
        pdata = le16_to_cpu(*((const __le16 *)&drift_info.chip->shadow[TMD2755_REG_PDATAL]));
    }
    else
    {
        pdata = (u16)((drift_info.chip->shadow[TMD2755_REG_PDATAL] << TMD2755_SHIFT_DRIFT_PDATA_L) |
                      (drift_info.chip->shadow[TMD2755_REG_PDATAH] & 0x3));
    }

    return(pdata);
}

static u16 acquire_temperature_reading(void)
{
    u16 temp_cnt;
    u8 prox_apc = drift_info.chip->params.prox_apc; /* save current apc value */

    /* Disable Prox and IRQ */
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Disabling Prox and IRQ...\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_ENABLE,  TMD2755_PEN,            0);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_INTENAB, TMD2755_INT_ST_PRX_IRQ, 0);

    /* Turn off APC for the temp reading - not part of HW state machine */
    /* Need to update data structure because acquire_current_pdata() uses it */
    drift_info.chip->params.prox_apc = PROX_APC_DISABLED;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_CFG6, TMD2755_MASK_APC, 
                   (drift_info.chip->params.prox_apc << TMD2755_SHIFT_APC));
                   
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Delaying %d ms for current prox cycle...\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, ADC_TEMP_DELAY_MS);

    mdelay(ADC_TEMP_DELAY_MS);

    /* Clear any pending Prox IRQ */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_STATUS,  TMD2755_INT_ST_PRX_IRQ, TMD2755_INT_ST_PRX_IRQ);

    /* Disable PWTIME */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_ENABLE, TMD2755_PWEN, 0);

    /* Disable HW Averaging */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                   TMD2755_REG_CALIBCFG, TMD2755_MASK_HW_AVG,
                   0);

    /* Enable Temp sensor, ADC mux for Temp, ADC Settle for Temp  */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_CFG1,
                   TMD2755_MASK_DRIFT_CFG1_EN_TEMP, (1 << TMD2755_SHIFT_DRIFT_CFG1_EN_TEMP));
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_DRFT_ADC_CFG,
                   TMD2755_MASK_DRIFT_ADC_SEL, TMD2755_ADC_SEL_TEMP);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_DRFT_ADC_CFG,
                   TMD2755_MASK_DRIFT_ADC_SETTLE,
                   TMD2755_ADC_SETTLE_TEMP << TMD2755_SHIFT_DRIFT_ADC_SETTLE);

    /* Start the temp sensor */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_DRFT_CALIB,
                   TMD2755_MASK_DRIFT_MAN_ADC_START,
                   TMD2755_ADC_MANUAL_START << TMD2755_SHIFT_DRIFT_MAN_ADC_START);

    /* Allow the ADC selector to settle, and read a temperature sample */
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Delaying %d ms to acquire temperature reading\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, PTIME_COMP_DELAY_MS);
    mdelay(PTIME_COMP_DELAY_MS);

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Performing a temperature reading\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);

    /* read pdata for the temperature count */
    temp_cnt = acquire_current_pdata();

    /* Switch ADC mux back to prox, disable temp sensor, reset ADC settling for prox */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_DRFT_ADC_CFG,
                   TMD2755_MASK_DRIFT_ADC_SEL, TMD2755_ADC_SEL_PROX);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_CFG1,
                   TMD2755_MASK_DRIFT_CFG1_EN_TEMP, (0 << TMD2755_SHIFT_DRIFT_CFG1_EN_TEMP));
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_DRFT_ADC_CFG,
                   TMD2755_MASK_DRIFT_ADC_SETTLE,
                   TMD2755_ADC_SETTLE_PROX << TMD2755_SHIFT_DRIFT_ADC_SETTLE);

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Re-enabling Prox and Irq...\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);

    /* Clear any pending interrupts and renable prox IRQ  */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_STATUS,  TMD2755_INT_ST_PRX_IRQ, TMD2755_INT_ST_PRX_IRQ);
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_INTENAB, TMD2755_INT_ST_PRX_IRQ, TMD2755_INT_ST_PRX_IRQ);

    /* Re-enable PWTIME */
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow,
                  TMD2755_REG_PWTIME, drift_info.chip->params.prox_wtime);

    /* Re enable HW Avg */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow,
                    TMD2755_REG_CALIBCFG, TMD2755_MASK_HW_AVG,
                   (drift_info.chip->params.prox_avg << TMD2755_SHIFT_HW_AVG));

    /* Restore apc to original setting */
    drift_info.chip->params.prox_apc = prox_apc;
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_CFG6, TMD2755_MASK_APC, 
                   (drift_info.chip->params.prox_apc << TMD2755_SHIFT_APC));
    
    /* Enable Prox Feature */
    ams_i2c_modify(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_ENABLE,  TMD2755_PEN, TMD2755_PEN);

    /* This delay is necessary so that prox can reacquire */
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Delaying %d ms to return to prox\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, ADC_PROX_DELAY_MS);
    mdelay(ADC_PROX_DELAY_MS);

    return(temp_cnt);
}

static s16 acquire_current_poffset(void)
{
    s16 poffset;
 
    ams_i2c_blk_read(drift_info.chip->client, TMD2755_REG_POFFSET_L, &drift_info.chip->shadow[TMD2755_REG_POFFSET_L], 2);
    poffset = drift_info.chip->shadow[TMD2755_REG_POFFSET_L];

    /* the high bit contains the sign */
    if (drift_info.chip->shadow[TMD2755_REG_POFFSET_H] & TMD2755_MASK_POFFSET_H)
    {
        poffset *= -1;
    }

    return(poffset);
}

static void update_current_poffset(s16 poffset)
{
    u8 sign_bit = 0, poffset_l = 0;

    /* If poffset is negative, write a 1 to the poffset_h register */
    if (poffset < 0)
    {
        sign_bit = 1;
        poffset_l = (u8)(poffset * (-1));
    }
    else
    {
        poffset_l = (u8)poffset;
    }

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Drift: Detect State:  Updating poffset on chip\n",
              MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
    /* The poffset is update based on a temperature reading and in the detect state */
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_POFFSET_L, poffset_l);
    ams_i2c_write(drift_info.chip->client, drift_info.chip->shadow, TMD2755_REG_POFFSET_H, sign_bit);

    return;
}

static bool update_drift_table(void)
{
    static u8 out_of_range = PDATA_OUT_OF_RANGE_INIT;
    s16 new_poffset = 0;
    u64 temp_pdata;
    bool ret_val = false;

    /* read the pdata - make sure its in the HW drift algo range */
    /* This range starts below DLL and extends above DUL1 - but below DUL2 */
    drift_log_info.pdata = acquire_current_pdata();

    /* if in release mode and pdata is not in the range DLL - x < pdata < DUL1 + x */
    /* where x is a constant, do not attempt to update table                       */
    if ((drift_log_info.pdata <= DRIFT_TABLE_PDATA_LOWER) || (drift_log_info.pdata >= DRIFT_TABLE_PDATA_UPPER))
    {
        out_of_range++;
        dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> pdata out of range: %d (%d,%d). out_of_range occurence = %d\n",
                 MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN,
                 __LINE__, drift_log_info.pdata, DRIFT_TABLE_PDATA_LOWER, DRIFT_TABLE_PDATA_UPPER, out_of_range);
    }
    else
    {
        out_of_range = PDATA_OUT_OF_RANGE_INIT;
    }

    if (out_of_range == PDATA_OUT_OF_RANGE_FIRST)
    {
        /* If detect out of range for first time, exit */
        return(ret_val);
    }

    /* read the poffset */
    drift_log_info.poffset = acquire_current_poffset();

    /* read the current temperature */
    drift_log_info.temp_cnt = acquire_temperature_reading();

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> prox pdata = %d (temp, poffset) = (%d, %d)\n",
              MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, drift_log_info.pdata,
              drift_log_info.temp_cnt, drift_log_info.poffset);

    /* Prevent any spurious value from updating the table */
    if ((drift_log_info.temp_cnt) > TEMP_CNT_MIN  && (drift_log_info.temp_cnt < TEMP_CNT_MAX))
    {
        /* If "two" consecutive pdata values are out of range, manually adjust the poffset */
        /* to cause the pdata to drop within range                                       */
        if (out_of_range == PDATA_OUT_OF_RANGE_MAX)
        {
            dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> pdata: [out of range: %d consecutive samples] - Manually adjust poffset\n",
                    MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, out_of_range);

            /* new_poffset = current_poffset + ((current_pdata - 30)/20) */
            /* 20 - normal counts per poffset change with ORE disabled */
            temp_pdata = drift_log_info.pdata - 30;
            do_div(temp_pdata, 20);
            new_poffset = temp_pdata + drift_log_info.poffset;

            /* update */
            dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> pdata=%d     previous poffset=%d    adjusted poffset=%d\n",
                    MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, drift_log_info.pdata, drift_log_info.poffset, new_poffset);
            update_current_poffset(new_poffset);
            drift_log_info.poffset = new_poffset;
            
            /* reset flags */
            out_of_range = PDATA_OUT_OF_RANGE_INIT;
        }
        set_drift_table(drift_log_info.temp_cnt, drift_log_info.poffset);
        ret_val = true;
    }
    else
    {
        dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> Invalid temperature count (%d), Ignoring\n",
              MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, drift_log_info.temp_cnt);
        ret_val = false;
    }

    return(ret_val);
}

static void update_poffset(void)
{
    /* read the current temperature */
    drift_log_info.temp_cnt = acquire_temperature_reading();

    /* retrieve the current poffset from the table */
    drift_log_info.poffset_table = get_drift_table_entry(drift_log_info.temp_cnt);

    dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> Drift: Detect State:  Temperature/Offset Pair from Table: (%d, %d)\n",
              MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__, drift_log_info.temp_cnt,
              drift_log_info.poffset_table);

    /* write poffset to device and read it back */
    update_current_poffset(drift_log_info.poffset_table);
    drift_log_info.poffset = acquire_current_poffset();

    /* read pdata for prox */
    drift_log_info.pdata = acquire_current_pdata();
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> PDATA during Drift Detect: %d\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
             drift_log_info.pdata);

    return;
}

static void dump_drift_table(void)
{
    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> tab1  %2d  %2d  %2d  %2d  %2d  %2d  %2d  %2d  %2d  %2d  %2d\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
             thermal_drift_table[0], thermal_drift_table[1], thermal_drift_table[2], thermal_drift_table[3],
             thermal_drift_table[4], thermal_drift_table[5], thermal_drift_table[6], thermal_drift_table[7],
             thermal_drift_table[8], thermal_drift_table[9], thermal_drift_table[10]);

    dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> -------------------------------------------------------------------------------------\n",
             MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);

    return;
}

static void drift_table_work(struct work_struct *work)
{
    u32 drift_table_period = msecs_to_jiffies(SEC_TO_MILLI * TIMER_PERIOD_SEC);

    AMS_MUTEX_LOCK(&drift_info.chip->lock);

    drift_log_info.state          = false;
    drift_log_info.pdata          = 0;
    drift_log_info.poffset        = 0;
    drift_log_info.temp_cnt       = 0;
    drift_log_info.poffset_table  = 0;
    drift_log_info.table_index    = 0;

    if (drift_info.chip->in_calib)
    {
        dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> Skipping temperature reading while in calibration.\n",
                MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
        schedule_delayed_work(&drift_info.drift_table_update, drift_table_period);
        AMS_MUTEX_UNLOCK(&drift_info.chip->lock);
        return;
    }

    /* if we are in the release state, attempt to update the thermal drift table */
    if (!drift_info.chip->prox_info.detected)
    {
        dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Drift: Release State: Updating drift table.\n",
                 MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
        drift_log_info.state = false;
        if (update_drift_table())
        {
            dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> log1: Detect State=%d Pdata=%d Poffset=%d Temp_ADC=%d Table Index=%d Poffset_Table=%d\n",
                    MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
                    drift_log_info.state, drift_log_info.pdata, drift_log_info.poffset, drift_log_info.temp_cnt, drift_log_info.table_index, drift_log_info.poffset_table);
        }
    }
    else /* must be in the detect state */
    {
        dev_info(&drift_info.chip->client->dev, "%*.*s():%*d --> Drift: Detect State: Updating poffset if necessary\n",
                 MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
        drift_log_info.state = true;
        update_poffset();
        dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> log1: Detect State=%d Pdata=%d Poffset=%d Temp_ADC=%d Table Index=%d Poffset_Table=%d\n",
                MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__,
                drift_log_info.state, drift_log_info.pdata, drift_log_info.poffset, drift_log_info.temp_cnt, drift_log_info.table_index, drift_log_info.poffset_table);
    }

    dump_drift_table();

    /* Schedule work for next iteration */
    schedule_delayed_work(&drift_info.drift_table_update, drift_table_period);

    AMS_MUTEX_UNLOCK(&drift_info.chip->lock);

    return;
}

/******************************************************************************/
/*                                                                            */
/*                             PUBLIC APIs                                    */
/*                                                                            */
/******************************************************************************/

/*
 * Custom chip configuration for use with Prototype
 */
void ams_tmd2755_sw_drift_setup(struct tmd2755_chip *chip)
{
    u32 drift_table_period = msecs_to_jiffies(SEC_TO_MILLI * TIMER_PERIOD_SEC);

    if (!drift_info.init)
    {
        drift_info.init = true;

        drift_info.chip = chip;

        /* update prox settings to custom values */
        custom_prox_config();
        init_drift_table(chip->params.poffset);

        INIT_DELAYED_WORK(&drift_info.drift_table_update, drift_table_work);
        schedule_delayed_work(&drift_info.drift_table_update, drift_table_period);
    }

    return;
}

void ams_tmd2755_sw_drift_shutdown(void)
{
    if (drift_info.init)
    {
        dev_err(&drift_info.chip->client->dev, "%*.*s():%*d --> Shutting down custom drift algorithm\n",
                MIN_KERNEL_LOG_LEN, MAX_KERNEL_LOG_LEN, __func__, LINE_NUM_KERNEL_LOG_LEN, __LINE__);
        cancel_delayed_work_sync(&drift_info.drift_table_update);
        drift_info.init = false;
        drift_info.chip = NULL;
        drift_info.table_modified = false;
    }
    return;
}

